from __future__ import annotations

from typing import TYPE_CHECKING, Literal, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field

from ._core import DomainModel, DomainModelApply, TypeApplyList, TypeList

if TYPE_CHECKING:
    from ._periods import PeriodsApply

__all__ = [
    "ReserveBidTimeSeries",
    "ReserveBidTimeSeriesApply",
    "ReserveBidTimeSeriesList",
    "ReserveBidTimeSeriesApplyList",
    "ReserveBidTimeSeriesFields",
    "ReserveBidTimeSeriesTextFields",
]


ReserveBidTimeSeriesTextFields = Literal["bid_type", "measure_unit", "currency", "direction_name", "reserve_object"]
ReserveBidTimeSeriesFields = Literal[
    "bid_type", "measure_unit", "currency", "price", "direction_name", "reserve_object"
]

_RESERVEBIDTIMESERIES_PROPERTIES_BY_FIELD = {
    "bid_type": "BidType",
    "measure_unit": "MeasureUnit",
    "currency": "Currency",
    "price": "Price",
    "direction_name": "DirectionName",
    "reserve_object": "ReserveObject",
}


class ReserveBidTimeSeries(DomainModel):
    space: str = "power-ops"
    bid_type: Optional[str] = Field(None, alias="BidType")
    measure_unit: Optional[str] = Field(None, alias="MeasureUnit")
    currency: Optional[str] = Field(None, alias="Currency")
    price: Optional[float] = Field(None, alias="Price")
    direction_name: Optional[str] = Field(None, alias="DirectionName")
    reserve_object: Optional[str] = Field(None, alias="ReserveObject")
    periods: Optional[list[str]] = Field(None, alias="Periods")

    def as_apply(self) -> ReserveBidTimeSeriesApply:
        return ReserveBidTimeSeriesApply(
            external_id=self.external_id,
            bid_type=self.bid_type,
            measure_unit=self.measure_unit,
            currency=self.currency,
            price=self.price,
            direction_name=self.direction_name,
            reserve_object=self.reserve_object,
            periods=self.periods,
        )


class ReserveBidTimeSeriesApply(DomainModelApply):
    space: str = "power-ops"
    bid_type: Optional[str] = Field(None, alias="BidType")
    measure_unit: Optional[str] = Field(None, alias="MeasureUnit")
    currency: Optional[str] = Field(None, alias="Currency")
    price: Optional[float] = Field(None, alias="Price")
    direction_name: Optional[str] = Field(None, alias="DirectionName")
    reserve_object: Optional[str] = Field(None, alias="ReserveObject")
    periods: Union[list[PeriodsApply], list[str], None] = Field(default=None, repr=False, alias="Periods")

    def _to_instances_apply(self, cache: set[str]) -> dm.InstancesApply:
        if self.external_id in cache:
            return dm.InstancesApply(dm.NodeApplyList([]), dm.EdgeApplyList([]))

        sources = []
        properties = {}
        if self.bid_type is not None:
            properties["BidType"] = self.bid_type
        if self.measure_unit is not None:
            properties["MeasureUnit"] = self.measure_unit
        if self.currency is not None:
            properties["Currency"] = self.currency
        if self.price is not None:
            properties["Price"] = self.price
        if self.direction_name is not None:
            properties["DirectionName"] = self.direction_name
        if self.reserve_object is not None:
            properties["ReserveObject"] = self.reserve_object
        if properties:
            source = dm.NodeOrEdgeData(
                source=dm.ContainerId("power-ops", "ReserveBidTimeSeries"),
                properties=properties,
            )
            sources.append(source)
        if sources:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.existing_version,
                sources=sources,
            )
            nodes = [this_node]
        else:
            nodes = []

        edges = []
        cache.add(self.external_id)

        for period in self.periods or []:
            edge = self._create_period_edge(period)
            if edge.external_id not in cache:
                edges.append(edge)
                cache.add(edge.external_id)

            if isinstance(period, DomainModelApply):
                instances = period._to_instances_apply(cache)
                nodes.extend(instances.nodes)
                edges.extend(instances.edges)

        return dm.InstancesApply(dm.NodeApplyList(nodes), dm.EdgeApplyList(edges))

    def _create_period_edge(self, period: Union[str, PeriodsApply]) -> dm.EdgeApply:
        if isinstance(period, str):
            end_node_ext_id = period
        elif isinstance(period, DomainModelApply):
            end_node_ext_id = period.external_id
        else:
            raise TypeError(f"Expected str or PeriodsApply, got {type(period)}")

        return dm.EdgeApply(
            space="power-ops",
            external_id=f"{self.external_id}:{end_node_ext_id}",
            type=dm.DirectRelationReference("power-ops", "ReserveBidTimeSeries.Periods"),
            start_node=dm.DirectRelationReference(self.space, self.external_id),
            end_node=dm.DirectRelationReference("power-ops", end_node_ext_id),
        )


class ReserveBidTimeSeriesList(TypeList[ReserveBidTimeSeries]):
    _NODE = ReserveBidTimeSeries

    def as_apply(self) -> ReserveBidTimeSeriesApplyList:
        return ReserveBidTimeSeriesApplyList([node.as_apply() for node in self.data])


class ReserveBidTimeSeriesApplyList(TypeApplyList[ReserveBidTimeSeriesApply]):
    _NODE = ReserveBidTimeSeriesApply



from __future__ import annotations

from typing import TYPE_CHECKING, Literal, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field

from ._core import DomainModel, DomainModelApply, TypeApplyList, TypeList

if TYPE_CHECKING:
    from ._plant import PlantApply
    from ._watercourse import WatercourseApply

__all__ = [
    "PriceArea",
    "PriceAreaApply",
    "PriceAreaList",
    "PriceAreaApplyList",
    "PriceAreaFields",
    "PriceAreaTextFields",
]


PriceAreaTextFields = Literal["name", "description", "dayahead_price_time_series"]
PriceAreaFields = Literal["name", "description", "dayahead_price_time_series"]

_PRICEAREA_PROPERTIES_BY_FIELD = {
    "name": "name",
    "description": "description",
    "dayahead_price_time_series": "dayaheadPriceTimeSeries",
}


class PriceArea(DomainModel):
    space: str = "power-ops"
    name: Optional[str] = None
    description: Optional[str] = None
    dayahead_price_time_series: Optional[str] = Field(None, alias="dayaheadPriceTimeSeries")
    plants: Optional[list[str]] = None
    watercourses: Optional[list[str]] = None

    def as_apply(self) -> PriceAreaApply:
        return PriceAreaApply(
            external_id=self.external_id,
            name=self.name,
            description=self.description,
            dayahead_price_time_series=self.dayahead_price_time_series,
            plants=self.plants,
            watercourses=self.watercourses,
        )


class PriceAreaApply(DomainModelApply):
    space: str = "power-ops"
    name: Optional[str] = None
    description: Optional[str] = None
    dayahead_price_time_series: Optional[str] = Field(None, alias="dayaheadPriceTimeSeries")
    plants: Union[list[PlantApply], list[str], None] = Field(default=None, repr=False)
    watercourses: Union[list[WatercourseApply], list[str], None] = Field(default=None, repr=False)

    def _to_instances_apply(self, cache: set[str]) -> dm.InstancesApply:
        if self.external_id in cache:
            return dm.InstancesApply(dm.NodeApplyList([]), dm.EdgeApplyList([]))

        sources = []
        properties = {}
        if self.name is not None:
            properties["name"] = self.name
        if self.description is not None:
            properties["description"] = self.description
        if self.dayahead_price_time_series is not None:
            properties["dayaheadPriceTimeSeries"] = self.dayahead_price_time_series
        if properties:
            source = dm.NodeOrEdgeData(
                source=dm.ContainerId("power-ops", "PriceArea"),
                properties=properties,
            )
            sources.append(source)
        if sources:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.existing_version,
                sources=sources,
            )
            nodes = [this_node]
        else:
            nodes = []

        edges = []
        cache.add(self.external_id)

        for plant in self.plants or []:
            edge = self._create_plant_edge(plant)
            if edge.external_id not in cache:
                edges.append(edge)
                cache.add(edge.external_id)

            if isinstance(plant, DomainModelApply):
                instances = plant._to_instances_apply(cache)
                nodes.extend(instances.nodes)
                edges.extend(instances.edges)

        for watercourse in self.watercourses or []:
            edge = self._create_watercourse_edge(watercourse)
            if edge.external_id not in cache:
                edges.append(edge)
                cache.add(edge.external_id)

            if isinstance(watercourse, DomainModelApply):
                instances = watercourse._to_instances_apply(cache)
                nodes.extend(instances.nodes)
                edges.extend(instances.edges)

        return dm.InstancesApply(dm.NodeApplyList(nodes), dm.EdgeApplyList(edges))

    def _create_plant_edge(self, plant: Union[str, PlantApply]) -> dm.EdgeApply:
        if isinstance(plant, str):
            end_node_ext_id = plant
        elif isinstance(plant, DomainModelApply):
            end_node_ext_id = plant.external_id
        else:
            raise TypeError(f"Expected str or PlantApply, got {type(plant)}")

        return dm.EdgeApply(
            space="power-ops",
            external_id=f"{self.external_id}:{end_node_ext_id}",
            type=dm.DirectRelationReference("power-ops", "PriceArea.plants"),
            start_node=dm.DirectRelationReference(self.space, self.external_id),
            end_node=dm.DirectRelationReference("power-ops", end_node_ext_id),
        )

    def _create_watercourse_edge(self, watercourse: Union[str, WatercourseApply]) -> dm.EdgeApply:
        if isinstance(watercourse, str):
            end_node_ext_id = watercourse
        elif isinstance(watercourse, DomainModelApply):
            end_node_ext_id = watercourse.external_id
        else:
            raise TypeError(f"Expected str or WatercourseApply, got {type(watercourse)}")

        return dm.EdgeApply(
            space="power-ops",
            external_id=f"{self.external_id}:{end_node_ext_id}",
            type=dm.DirectRelationReference("power-ops", "PriceArea.watercourses"),
            start_node=dm.DirectRelationReference(self.space, self.external_id),
            end_node=dm.DirectRelationReference("power-ops", end_node_ext_id),
        )


class PriceAreaList(TypeList[PriceArea]):
    _NODE = PriceArea

    def as_apply(self) -> PriceAreaApplyList:
        return PriceAreaApplyList([node.as_apply() for node in self.data])


class PriceAreaApplyList(TypeApplyList[PriceAreaApply]):
    _NODE = PriceAreaApply



from __future__ import annotations

from collections.abc import Sequence
from typing import overload

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.data_classes.data_modeling.instances import InstanceAggregationResultList

from cognite.powerops.client._generated.data_classes import (
    BidTimeSeries,
    BidTimeSeriesApply,
    BidTimeSeriesApplyList,
    BidTimeSeriesFields,
    BidTimeSeriesList,
    BidTimeSeriesTextFields,
)
from cognite.powerops.client._generated.data_classes._bid_time_series import _BIDTIMESERIES_PROPERTIES_BY_FIELD

from ._core import DEFAULT_LIMIT_READ, Aggregations, TypeAPI


class BidTimeSeriesAPI(TypeAPI[BidTimeSeries, BidTimeSeriesApply, BidTimeSeriesList]):
    def __init__(self, client: CogniteClient, view_id: dm.ViewId):
        super().__init__(
            client=client,
            sources=view_id,
            class_type=BidTimeSeries,
            class_apply_type=BidTimeSeriesApply,
            class_list=BidTimeSeriesList,
        )
        self._view_id = view_id

    def apply(
        self, bid_time_series: BidTimeSeriesApply | Sequence[BidTimeSeriesApply], replace: bool = False
    ) -> dm.InstancesApplyResult:
        if isinstance(bid_time_series, BidTimeSeriesApply):
            instances = bid_time_series.to_instances_apply()
        else:
            instances = BidTimeSeriesApplyList(bid_time_series).to_instances_apply()
        return self._client.data_modeling.instances.apply(
            nodes=instances.nodes,
            edges=instances.edges,
            auto_create_start_nodes=True,
            auto_create_end_nodes=True,
            replace=replace,
        )

    def delete(self, external_id: str | Sequence[str], space="power-ops") -> dm.InstancesDeleteResult:
        if isinstance(external_id, str):
            return self._client.data_modeling.instances.delete(nodes=(space, external_id))
        else:
            return self._client.data_modeling.instances.delete(
                nodes=[(space, id) for id in external_id],
            )

    @overload
    def retrieve(self, external_id: str) -> BidTimeSeries:
        ...

    @overload
    def retrieve(self, external_id: Sequence[str]) -> BidTimeSeriesList:
        ...

    def retrieve(self, external_id: str | Sequence[str]) -> BidTimeSeries | BidTimeSeriesList:
        if isinstance(external_id, str):
            return self._retrieve((self._sources.space, external_id))
        else:
            return self._retrieve([(self._sources.space, ext_id) for ext_id in external_id])

    def search(
        self,
        query: str,
        properties: BidTimeSeriesTextFields | Sequence[BidTimeSeriesTextFields] | None = None,
        m_rid: str | list[str] | None = None,
        m_rid_prefix: str | None = None,
        auction: str | list[str] | None = None,
        auction_prefix: str | None = None,
        acquiring_domain: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        connecting_domain: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        provider_market_participant: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        quantity_measure_unit_name: str | list[str] | None = None,
        quantity_measure_unit_name_prefix: str | None = None,
        currency_unit_name: str | list[str] | None = None,
        currency_unit_name_prefix: str | None = None,
        price_measure_unit_name: str | list[str] | None = None,
        price_measure_unit_name_prefix: str | None = None,
        divisible: bool | None = None,
        linked_bid: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        multipart_bid: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        exclusive_bid: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        block_bid: bool | None = None,
        min_status: int | None = None,
        max_status: int | None = None,
        min_priority: int | None = None,
        max_priority: int | None = None,
        registered_resources_mrid: str | list[str] | None = None,
        registered_resources_mrid_prefix: str | None = None,
        flow_direction: str | list[str] | None = None,
        flow_direction_prefix: str | None = None,
        min_step_increment_quantity: float | None = None,
        max_step_increment_quantity: float | None = None,
        energy_price_measure_unit: str | list[str] | None = None,
        energy_price_measure_unit_prefix: str | None = None,
        market_agreement: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        activation_constraint: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        resting_constraint: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        minimum_constraint: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        maximum_constraint: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        standard_market_product_type: str | list[str] | None = None,
        standard_market_product_type_prefix: str | None = None,
        original_market_product_type: str | list[str] | None = None,
        original_market_product_type_prefix: str | None = None,
        validity_period: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        reason: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> BidTimeSeriesList:
        filter_ = _create_filter(
            self._view_id,
            m_rid,
            m_rid_prefix,
            auction,
            auction_prefix,
            acquiring_domain,
            connecting_domain,
            provider_market_participant,
            quantity_measure_unit_name,
            quantity_measure_unit_name_prefix,
            currency_unit_name,
            currency_unit_name_prefix,
            price_measure_unit_name,
            price_measure_unit_name_prefix,
            divisible,
            linked_bid,
            multipart_bid,
            exclusive_bid,
            block_bid,
            min_status,
            max_status,
            min_priority,
            max_priority,
            registered_resources_mrid,
            registered_resources_mrid_prefix,
            flow_direction,
            flow_direction_prefix,
            min_step_increment_quantity,
            max_step_increment_quantity,
            energy_price_measure_unit,
            energy_price_measure_unit_prefix,
            market_agreement,
            activation_constraint,
            resting_constraint,
            minimum_constraint,
            maximum_constraint,
            standard_market_product_type,
            standard_market_product_type_prefix,
            original_market_product_type,
            original_market_product_type_prefix,
            validity_period,
            reason,
            external_id_prefix,
            filter,
        )
        return self._search(self._view_id, query, _BIDTIMESERIES_PROPERTIES_BY_FIELD, properties, filter_, limit)

    @overload
    def aggregate(
        self,
        aggregations: Aggregations
        | dm.aggregations.MetricAggregation
        | Sequence[Aggregations]
        | Sequence[dm.aggregations.MetricAggregation],
        property: BidTimeSeriesFields | Sequence[BidTimeSeriesFields] | None = None,
        group_by: None = None,
        query: str | None = None,
        search_properties: BidTimeSeriesTextFields | Sequence[BidTimeSeriesTextFields] | None = None,
        m_rid: str | list[str] | None = None,
        m_rid_prefix: str | None = None,
        auction: str | list[str] | None = None,
        auction_prefix: str | None = None,
        acquiring_domain: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        connecting_domain: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        provider_market_participant: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        quantity_measure_unit_name: str | list[str] | None = None,
        quantity_measure_unit_name_prefix: str | None = None,
        currency_unit_name: str | list[str] | None = None,
        currency_unit_name_prefix: str | None = None,
        price_measure_unit_name: str | list[str] | None = None,
        price_measure_unit_name_prefix: str | None = None,
        divisible: bool | None = None,
        linked_bid: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        multipart_bid: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        exclusive_bid: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        block_bid: bool | None = None,
        min_status: int | None = None,
        max_status: int | None = None,
        min_priority: int | None = None,
        max_priority: int | None = None,
        registered_resources_mrid: str | list[str] | None = None,
        registered_resources_mrid_prefix: str | None = None,
        flow_direction: str | list[str] | None = None,
        flow_direction_prefix: str | None = None,
        min_step_increment_quantity: float | None = None,
        max_step_increment_quantity: float | None = None,
        energy_price_measure_unit: str | list[str] | None = None,
        energy_price_measure_unit_prefix: str | None = None,
        market_agreement: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        activation_constraint: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        resting_constraint: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        minimum_constraint: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        maximum_constraint: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        standard_market_product_type: str | list[str] | None = None,
        standard_market_product_type_prefix: str | None = None,
        original_market_product_type: str | list[str] | None = None,
        original_market_product_type_prefix: str | None = None,
        validity_period: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        reason: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue]:
        ...

    @overload
    def aggregate(
        self,
        aggregations: Aggregations
        | dm.aggregations.MetricAggregation
        | Sequence[Aggregations]
        | Sequence[dm.aggregations.MetricAggregation],
        property: BidTimeSeriesFields | Sequence[BidTimeSeriesFields] | None = None,
        group_by: BidTimeSeriesFields | Sequence[BidTimeSeriesFields] = None,
        query: str | None = None,
        search_properties: BidTimeSeriesTextFields | Sequence[BidTimeSeriesTextFields] | None = None,
        m_rid: str | list[str] | None = None,
        m_rid_prefix: str | None = None,
        auction: str | list[str] | None = None,
        auction_prefix: str | None = None,
        acquiring_domain: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        connecting_domain: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        provider_market_participant: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        quantity_measure_unit_name: str | list[str] | None = None,
        quantity_measure_unit_name_prefix: str | None = None,
        currency_unit_name: str | list[str] | None = None,
        currency_unit_name_prefix: str | None = None,
        price_measure_unit_name: str | list[str] | None = None,
        price_measure_unit_name_prefix: str | None = None,
        divisible: bool | None = None,
        linked_bid: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        multipart_bid: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        exclusive_bid: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        block_bid: bool | None = None,
        min_status: int | None = None,
        max_status: int | None = None,
        min_priority: int | None = None,
        max_priority: int | None = None,
        registered_resources_mrid: str | list[str] | None = None,
        registered_resources_mrid_prefix: str | None = None,
        flow_direction: str | list[str] | None = None,
        flow_direction_prefix: str | None = None,
        min_step_increment_quantity: float | None = None,
        max_step_increment_quantity: float | None = None,
        energy_price_measure_unit: str | list[str] | None = None,
        energy_price_measure_unit_prefix: str | None = None,
        market_agreement: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        activation_constraint: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        resting_constraint: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        minimum_constraint: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        maximum_constraint: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        standard_market_product_type: str | list[str] | None = None,
        standard_market_product_type_prefix: str | None = None,
        original_market_product_type: str | list[str] | None = None,
        original_market_product_type_prefix: str | None = None,
        validity_period: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        reason: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> InstanceAggregationResultList:
        ...

    def aggregate(
        self,
        aggregate: Aggregations
        | dm.aggregations.MetricAggregation
        | Sequence[Aggregations]
        | Sequence[dm.aggregations.MetricAggregation],
        property: BidTimeSeriesFields | Sequence[BidTimeSeriesFields] | None = None,
        group_by: BidTimeSeriesFields | Sequence[BidTimeSeriesFields] | None = None,
        query: str | None = None,
        search_property: BidTimeSeriesTextFields | Sequence[BidTimeSeriesTextFields] | None = None,
        m_rid: str | list[str] | None = None,
        m_rid_prefix: str | None = None,
        auction: str | list[str] | None = None,
        auction_prefix: str | None = None,
        acquiring_domain: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        connecting_domain: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        provider_market_participant: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        quantity_measure_unit_name: str | list[str] | None = None,
        quantity_measure_unit_name_prefix: str | None = None,
        currency_unit_name: str | list[str] | None = None,
        currency_unit_name_prefix: str | None = None,
        price_measure_unit_name: str | list[str] | None = None,
        price_measure_unit_name_prefix: str | None = None,
        divisible: bool | None = None,
        linked_bid: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        multipart_bid: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        exclusive_bid: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        block_bid: bool | None = None,
        min_status: int | None = None,
        max_status: int | None = None,
        min_priority: int | None = None,
        max_priority: int | None = None,
        registered_resources_mrid: str | list[str] | None = None,
        registered_resources_mrid_prefix: str | None = None,
        flow_direction: str | list[str] | None = None,
        flow_direction_prefix: str | None = None,
        min_step_increment_quantity: float | None = None,
        max_step_increment_quantity: float | None = None,
        energy_price_measure_unit: str | list[str] | None = None,
        energy_price_measure_unit_prefix: str | None = None,
        market_agreement: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        activation_constraint: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        resting_constraint: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        minimum_constraint: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        maximum_constraint: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        standard_market_product_type: str | list[str] | None = None,
        standard_market_product_type_prefix: str | None = None,
        original_market_product_type: str | list[str] | None = None,
        original_market_product_type_prefix: str | None = None,
        validity_period: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        reason: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue] | InstanceAggregationResultList:
        filter_ = _create_filter(
            self._view_id,
            m_rid,
            m_rid_prefix,
            auction,
            auction_prefix,
            acquiring_domain,
            connecting_domain,
            provider_market_participant,
            quantity_measure_unit_name,
            quantity_measure_unit_name_prefix,
            currency_unit_name,
            currency_unit_name_prefix,
            price_measure_unit_name,
            price_measure_unit_name_prefix,
            divisible,
            linked_bid,
            multipart_bid,
            exclusive_bid,
            block_bid,
            min_status,
            max_status,
            min_priority,
            max_priority,
            registered_resources_mrid,
            registered_resources_mrid_prefix,
            flow_direction,
            flow_direction_prefix,
            min_step_increment_quantity,
            max_step_increment_quantity,
            energy_price_measure_unit,
            energy_price_measure_unit_prefix,
            market_agreement,
            activation_constraint,
            resting_constraint,
            minimum_constraint,
            maximum_constraint,
            standard_market_product_type,
            standard_market_product_type_prefix,
            original_market_product_type,
            original_market_product_type_prefix,
            validity_period,
            reason,
            external_id_prefix,
            filter,
        )
        return self._aggregate(
            self._view_id,
            aggregate,
            _BIDTIMESERIES_PROPERTIES_BY_FIELD,
            property,
            group_by,
            query,
            search_property,
            limit,
            filter_,
        )

    def histogram(
        self,
        property: BidTimeSeriesFields,
        interval: float,
        query: str | None = None,
        search_property: BidTimeSeriesTextFields | Sequence[BidTimeSeriesTextFields] | None = None,
        m_rid: str | list[str] | None = None,
        m_rid_prefix: str | None = None,
        auction: str | list[str] | None = None,
        auction_prefix: str | None = None,
        acquiring_domain: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        connecting_domain: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        provider_market_participant: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        quantity_measure_unit_name: str | list[str] | None = None,
        quantity_measure_unit_name_prefix: str | None = None,
        currency_unit_name: str | list[str] | None = None,
        currency_unit_name_prefix: str | None = None,
        price_measure_unit_name: str | list[str] | None = None,
        price_measure_unit_name_prefix: str | None = None,
        divisible: bool | None = None,
        linked_bid: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        multipart_bid: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        exclusive_bid: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        block_bid: bool | None = None,
        min_status: int | None = None,
        max_status: int | None = None,
        min_priority: int | None = None,
        max_priority: int | None = None,
        registered_resources_mrid: str | list[str] | None = None,
        registered_resources_mrid_prefix: str | None = None,
        flow_direction: str | list[str] | None = None,
        flow_direction_prefix: str | None = None,
        min_step_increment_quantity: float | None = None,
        max_step_increment_quantity: float | None = None,
        energy_price_measure_unit: str | list[str] | None = None,
        energy_price_measure_unit_prefix: str | None = None,
        market_agreement: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        activation_constraint: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        resting_constraint: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        minimum_constraint: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        maximum_constraint: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        standard_market_product_type: str | list[str] | None = None,
        standard_market_product_type_prefix: str | None = None,
        original_market_product_type: str | list[str] | None = None,
        original_market_product_type_prefix: str | None = None,
        validity_period: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        reason: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> dm.aggregations.HistogramValue:
        filter_ = _create_filter(
            self._view_id,
            m_rid,
            m_rid_prefix,
            auction,
            auction_prefix,
            acquiring_domain,
            connecting_domain,
            provider_market_participant,
            quantity_measure_unit_name,
            quantity_measure_unit_name_prefix,
            currency_unit_name,
            currency_unit_name_prefix,
            price_measure_unit_name,
            price_measure_unit_name_prefix,
            divisible,
            linked_bid,
            multipart_bid,
            exclusive_bid,
            block_bid,
            min_status,
            max_status,
            min_priority,
            max_priority,
            registered_resources_mrid,
            registered_resources_mrid_prefix,
            flow_direction,
            flow_direction_prefix,
            min_step_increment_quantity,
            max_step_increment_quantity,
            energy_price_measure_unit,
            energy_price_measure_unit_prefix,
            market_agreement,
            activation_constraint,
            resting_constraint,
            minimum_constraint,
            maximum_constraint,
            standard_market_product_type,
            standard_market_product_type_prefix,
            original_market_product_type,
            original_market_product_type_prefix,
            validity_period,
            reason,
            external_id_prefix,
            filter,
        )
        return self._histogram(
            self._view_id,
            property,
            interval,
            _BIDTIMESERIES_PROPERTIES_BY_FIELD,
            query,
            search_property,
            limit,
            filter_,
        )

    def list(
        self,
        m_rid: str | list[str] | None = None,
        m_rid_prefix: str | None = None,
        auction: str | list[str] | None = None,
        auction_prefix: str | None = None,
        acquiring_domain: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        connecting_domain: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        provider_market_participant: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        quantity_measure_unit_name: str | list[str] | None = None,
        quantity_measure_unit_name_prefix: str | None = None,
        currency_unit_name: str | list[str] | None = None,
        currency_unit_name_prefix: str | None = None,
        price_measure_unit_name: str | list[str] | None = None,
        price_measure_unit_name_prefix: str | None = None,
        divisible: bool | None = None,
        linked_bid: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        multipart_bid: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        exclusive_bid: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        block_bid: bool | None = None,
        min_status: int | None = None,
        max_status: int | None = None,
        min_priority: int | None = None,
        max_priority: int | None = None,
        registered_resources_mrid: str | list[str] | None = None,
        registered_resources_mrid_prefix: str | None = None,
        flow_direction: str | list[str] | None = None,
        flow_direction_prefix: str | None = None,
        min_step_increment_quantity: float | None = None,
        max_step_increment_quantity: float | None = None,
        energy_price_measure_unit: str | list[str] | None = None,
        energy_price_measure_unit_prefix: str | None = None,
        market_agreement: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        activation_constraint: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        resting_constraint: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        minimum_constraint: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        maximum_constraint: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        standard_market_product_type: str | list[str] | None = None,
        standard_market_product_type_prefix: str | None = None,
        original_market_product_type: str | list[str] | None = None,
        original_market_product_type_prefix: str | None = None,
        validity_period: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        reason: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> BidTimeSeriesList:
        filter_ = _create_filter(
            self._view_id,
            m_rid,
            m_rid_prefix,
            auction,
            auction_prefix,
            acquiring_domain,
            connecting_domain,
            provider_market_participant,
            quantity_measure_unit_name,
            quantity_measure_unit_name_prefix,
            currency_unit_name,
            currency_unit_name_prefix,
            price_measure_unit_name,
            price_measure_unit_name_prefix,
            divisible,
            linked_bid,
            multipart_bid,
            exclusive_bid,
            block_bid,
            min_status,
            max_status,
            min_priority,
            max_priority,
            registered_resources_mrid,
            registered_resources_mrid_prefix,
            flow_direction,
            flow_direction_prefix,
            min_step_increment_quantity,
            max_step_increment_quantity,
            energy_price_measure_unit,
            energy_price_measure_unit_prefix,
            market_agreement,
            activation_constraint,
            resting_constraint,
            minimum_constraint,
            maximum_constraint,
            standard_market_product_type,
            standard_market_product_type_prefix,
            original_market_product_type,
            original_market_product_type_prefix,
            validity_period,
            reason,
            external_id_prefix,
            filter,
        )

        return self._list(limit=limit, filter=filter_)


def _create_filter(
    view_id: dm.ViewId,
    m_rid: str | list[str] | None = None,
    m_rid_prefix: str | None = None,
    auction: str | list[str] | None = None,
    auction_prefix: str | None = None,
    acquiring_domain: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    connecting_domain: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    provider_market_participant: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    quantity_measure_unit_name: str | list[str] | None = None,
    quantity_measure_unit_name_prefix: str | None = None,
    currency_unit_name: str | list[str] | None = None,
    currency_unit_name_prefix: str | None = None,
    price_measure_unit_name: str | list[str] | None = None,
    price_measure_unit_name_prefix: str | None = None,
    divisible: bool | None = None,
    linked_bid: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    multipart_bid: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    exclusive_bid: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    block_bid: bool | None = None,
    min_status: int | None = None,
    max_status: int | None = None,
    min_priority: int | None = None,
    max_priority: int | None = None,
    registered_resources_mrid: str | list[str] | None = None,
    registered_resources_mrid_prefix: str | None = None,
    flow_direction: str | list[str] | None = None,
    flow_direction_prefix: str | None = None,
    min_step_increment_quantity: float | None = None,
    max_step_increment_quantity: float | None = None,
    energy_price_measure_unit: str | list[str] | None = None,
    energy_price_measure_unit_prefix: str | None = None,
    market_agreement: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    activation_constraint: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    resting_constraint: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    minimum_constraint: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    maximum_constraint: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    standard_market_product_type: str | list[str] | None = None,
    standard_market_product_type_prefix: str | None = None,
    original_market_product_type: str | list[str] | None = None,
    original_market_product_type_prefix: str | None = None,
    validity_period: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    reason: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    external_id_prefix: str | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if m_rid and isinstance(m_rid, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("mRID"), value=m_rid))
    if m_rid and isinstance(m_rid, list):
        filters.append(dm.filters.In(view_id.as_property_ref("mRID"), values=m_rid))
    if m_rid_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("mRID"), value=m_rid_prefix))
    if auction and isinstance(auction, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("auction"), value=auction))
    if auction and isinstance(auction, list):
        filters.append(dm.filters.In(view_id.as_property_ref("auction"), values=auction))
    if auction_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("auction"), value=auction_prefix))
    if acquiring_domain and isinstance(acquiring_domain, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("acquiringDomain"), value={"space": "power-ops", "externalId": acquiring_domain}
            )
        )
    if acquiring_domain and isinstance(acquiring_domain, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("acquiringDomain"),
                value={"space": acquiring_domain[0], "externalId": acquiring_domain[1]},
            )
        )
    if acquiring_domain and isinstance(acquiring_domain, list) and isinstance(acquiring_domain[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("acquiringDomain"),
                values=[{"space": "power-ops", "externalId": item} for item in acquiring_domain],
            )
        )
    if acquiring_domain and isinstance(acquiring_domain, list) and isinstance(acquiring_domain[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("acquiringDomain"),
                values=[{"space": item[0], "externalId": item[1]} for item in acquiring_domain],
            )
        )
    if connecting_domain and isinstance(connecting_domain, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("connectingDomain"),
                value={"space": "power-ops", "externalId": connecting_domain},
            )
        )
    if connecting_domain and isinstance(connecting_domain, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("connectingDomain"),
                value={"space": connecting_domain[0], "externalId": connecting_domain[1]},
            )
        )
    if connecting_domain and isinstance(connecting_domain, list) and isinstance(connecting_domain[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("connectingDomain"),
                values=[{"space": "power-ops", "externalId": item} for item in connecting_domain],
            )
        )
    if connecting_domain and isinstance(connecting_domain, list) and isinstance(connecting_domain[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("connectingDomain"),
                values=[{"space": item[0], "externalId": item[1]} for item in connecting_domain],
            )
        )
    if provider_market_participant and isinstance(provider_market_participant, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("providerMarketParticipant"),
                value={"space": "power-ops", "externalId": provider_market_participant},
            )
        )
    if provider_market_participant and isinstance(provider_market_participant, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("providerMarketParticipant"),
                value={"space": provider_market_participant[0], "externalId": provider_market_participant[1]},
            )
        )
    if (
        provider_market_participant
        and isinstance(provider_market_participant, list)
        and isinstance(provider_market_participant[0], str)
    ):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("providerMarketParticipant"),
                values=[{"space": "power-ops", "externalId": item} for item in provider_market_participant],
            )
        )
    if (
        provider_market_participant
        and isinstance(provider_market_participant, list)
        and isinstance(provider_market_participant[0], tuple)
    ):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("providerMarketParticipant"),
                values=[{"space": item[0], "externalId": item[1]} for item in provider_market_participant],
            )
        )
    if quantity_measure_unit_name and isinstance(quantity_measure_unit_name, str):
        filters.append(
            dm.filters.Equals(view_id.as_property_ref("quantityMeasureUnitName"), value=quantity_measure_unit_name)
        )
    if quantity_measure_unit_name and isinstance(quantity_measure_unit_name, list):
        filters.append(
            dm.filters.In(view_id.as_property_ref("quantityMeasureUnitName"), values=quantity_measure_unit_name)
        )
    if quantity_measure_unit_name_prefix:
        filters.append(
            dm.filters.Prefix(
                view_id.as_property_ref("quantityMeasureUnitName"), value=quantity_measure_unit_name_prefix
            )
        )
    if currency_unit_name and isinstance(currency_unit_name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("currencyUnitName"), value=currency_unit_name))
    if currency_unit_name and isinstance(currency_unit_name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("currencyUnitName"), values=currency_unit_name))
    if currency_unit_name_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("currencyUnitName"), value=currency_unit_name_prefix))
    if price_measure_unit_name and isinstance(price_measure_unit_name, str):
        filters.append(
            dm.filters.Equals(view_id.as_property_ref("priceMeasureUnitName"), value=price_measure_unit_name)
        )
    if price_measure_unit_name and isinstance(price_measure_unit_name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("priceMeasureUnitName"), values=price_measure_unit_name))
    if price_measure_unit_name_prefix:
        filters.append(
            dm.filters.Prefix(view_id.as_property_ref("priceMeasureUnitName"), value=price_measure_unit_name_prefix)
        )
    if divisible and isinstance(divisible, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("divisible"), value=divisible))
    if linked_bid and isinstance(linked_bid, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("linkedBid"), value={"space": "power-ops", "externalId": linked_bid}
            )
        )
    if linked_bid and isinstance(linked_bid, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("linkedBid"), value={"space": linked_bid[0], "externalId": linked_bid[1]}
            )
        )
    if linked_bid and isinstance(linked_bid, list) and isinstance(linked_bid[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("linkedBid"),
                values=[{"space": "power-ops", "externalId": item} for item in linked_bid],
            )
        )
    if linked_bid and isinstance(linked_bid, list) and isinstance(linked_bid[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("linkedBid"),
                values=[{"space": item[0], "externalId": item[1]} for item in linked_bid],
            )
        )
    if multipart_bid and isinstance(multipart_bid, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("multipartBid"), value={"space": "power-ops", "externalId": multipart_bid}
            )
        )
    if multipart_bid and isinstance(multipart_bid, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("multipartBid"),
                value={"space": multipart_bid[0], "externalId": multipart_bid[1]},
            )
        )
    if multipart_bid and isinstance(multipart_bid, list) and isinstance(multipart_bid[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("multipartBid"),
                values=[{"space": "power-ops", "externalId": item} for item in multipart_bid],
            )
        )
    if multipart_bid and isinstance(multipart_bid, list) and isinstance(multipart_bid[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("multipartBid"),
                values=[{"space": item[0], "externalId": item[1]} for item in multipart_bid],
            )
        )
    if exclusive_bid and isinstance(exclusive_bid, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("exclusiveBid"), value={"space": "power-ops", "externalId": exclusive_bid}
            )
        )
    if exclusive_bid and isinstance(exclusive_bid, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("exclusiveBid"),
                value={"space": exclusive_bid[0], "externalId": exclusive_bid[1]},
            )
        )
    if exclusive_bid and isinstance(exclusive_bid, list) and isinstance(exclusive_bid[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("exclusiveBid"),
                values=[{"space": "power-ops", "externalId": item} for item in exclusive_bid],
            )
        )
    if exclusive_bid and isinstance(exclusive_bid, list) and isinstance(exclusive_bid[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("exclusiveBid"),
                values=[{"space": item[0], "externalId": item[1]} for item in exclusive_bid],
            )
        )
    if block_bid and isinstance(block_bid, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("blockBid"), value=block_bid))
    if min_status or max_status:
        filters.append(dm.filters.Range(view_id.as_property_ref("status"), gte=min_status, lte=max_status))
    if min_priority or max_priority:
        filters.append(dm.filters.Range(view_id.as_property_ref("priority"), gte=min_priority, lte=max_priority))
    if registered_resources_mrid and isinstance(registered_resources_mrid, str):
        filters.append(
            dm.filters.Equals(view_id.as_property_ref("registeredResourcesMRID"), value=registered_resources_mrid)
        )
    if registered_resources_mrid and isinstance(registered_resources_mrid, list):
        filters.append(
            dm.filters.In(view_id.as_property_ref("registeredResourcesMRID"), values=registered_resources_mrid)
        )
    if registered_resources_mrid_prefix:
        filters.append(
            dm.filters.Prefix(
                view_id.as_property_ref("registeredResourcesMRID"), value=registered_resources_mrid_prefix
            )
        )
    if flow_direction and isinstance(flow_direction, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("flowDirection"), value=flow_direction))
    if flow_direction and isinstance(flow_direction, list):
        filters.append(dm.filters.In(view_id.as_property_ref("flowDirection"), values=flow_direction))
    if flow_direction_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("flowDirection"), value=flow_direction_prefix))
    if min_step_increment_quantity or max_step_increment_quantity:
        filters.append(
            dm.filters.Range(
                view_id.as_property_ref("stepIncrementQuantity"),
                gte=min_step_increment_quantity,
                lte=max_step_increment_quantity,
            )
        )
    if energy_price_measure_unit and isinstance(energy_price_measure_unit, str):
        filters.append(
            dm.filters.Equals(view_id.as_property_ref("energyPriceMeasureUnit"), value=energy_price_measure_unit)
        )
    if energy_price_measure_unit and isinstance(energy_price_measure_unit, list):
        filters.append(
            dm.filters.In(view_id.as_property_ref("energyPriceMeasureUnit"), values=energy_price_measure_unit)
        )
    if energy_price_measure_unit_prefix:
        filters.append(
            dm.filters.Prefix(view_id.as_property_ref("energyPriceMeasureUnit"), value=energy_price_measure_unit_prefix)
        )
    if market_agreement and isinstance(market_agreement, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("marketAgreement"), value={"space": "power-ops", "externalId": market_agreement}
            )
        )
    if market_agreement and isinstance(market_agreement, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("marketAgreement"),
                value={"space": market_agreement[0], "externalId": market_agreement[1]},
            )
        )
    if market_agreement and isinstance(market_agreement, list) and isinstance(market_agreement[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("marketAgreement"),
                values=[{"space": "power-ops", "externalId": item} for item in market_agreement],
            )
        )
    if market_agreement and isinstance(market_agreement, list) and isinstance(market_agreement[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("marketAgreement"),
                values=[{"space": item[0], "externalId": item[1]} for item in market_agreement],
            )
        )
    if activation_constraint and isinstance(activation_constraint, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("activationConstraint"),
                value={"space": "power-ops", "externalId": activation_constraint},
            )
        )
    if activation_constraint and isinstance(activation_constraint, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("activationConstraint"),
                value={"space": activation_constraint[0], "externalId": activation_constraint[1]},
            )
        )
    if activation_constraint and isinstance(activation_constraint, list) and isinstance(activation_constraint[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("activationConstraint"),
                values=[{"space": "power-ops", "externalId": item} for item in activation_constraint],
            )
        )
    if (
        activation_constraint
        and isinstance(activation_constraint, list)
        and isinstance(activation_constraint[0], tuple)
    ):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("activationConstraint"),
                values=[{"space": item[0], "externalId": item[1]} for item in activation_constraint],
            )
        )
    if resting_constraint and isinstance(resting_constraint, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("restingConstraint"),
                value={"space": "power-ops", "externalId": resting_constraint},
            )
        )
    if resting_constraint and isinstance(resting_constraint, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("restingConstraint"),
                value={"space": resting_constraint[0], "externalId": resting_constraint[1]},
            )
        )
    if resting_constraint and isinstance(resting_constraint, list) and isinstance(resting_constraint[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("restingConstraint"),
                values=[{"space": "power-ops", "externalId": item} for item in resting_constraint],
            )
        )
    if resting_constraint and isinstance(resting_constraint, list) and isinstance(resting_constraint[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("restingConstraint"),
                values=[{"space": item[0], "externalId": item[1]} for item in resting_constraint],
            )
        )
    if minimum_constraint and isinstance(minimum_constraint, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("minimumConstraint"),
                value={"space": "power-ops", "externalId": minimum_constraint},
            )
        )
    if minimum_constraint and isinstance(minimum_constraint, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("minimumConstraint"),
                value={"space": minimum_constraint[0], "externalId": minimum_constraint[1]},
            )
        )
    if minimum_constraint and isinstance(minimum_constraint, list) and isinstance(minimum_constraint[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("minimumConstraint"),
                values=[{"space": "power-ops", "externalId": item} for item in minimum_constraint],
            )
        )
    if minimum_constraint and isinstance(minimum_constraint, list) and isinstance(minimum_constraint[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("minimumConstraint"),
                values=[{"space": item[0], "externalId": item[1]} for item in minimum_constraint],
            )
        )
    if maximum_constraint and isinstance(maximum_constraint, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("maximumConstraint"),
                value={"space": "power-ops", "externalId": maximum_constraint},
            )
        )
    if maximum_constraint and isinstance(maximum_constraint, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("maximumConstraint"),
                value={"space": maximum_constraint[0], "externalId": maximum_constraint[1]},
            )
        )
    if maximum_constraint and isinstance(maximum_constraint, list) and isinstance(maximum_constraint[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("maximumConstraint"),
                values=[{"space": "power-ops", "externalId": item} for item in maximum_constraint],
            )
        )
    if maximum_constraint and isinstance(maximum_constraint, list) and isinstance(maximum_constraint[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("maximumConstraint"),
                values=[{"space": item[0], "externalId": item[1]} for item in maximum_constraint],
            )
        )
    if standard_market_product_type and isinstance(standard_market_product_type, str):
        filters.append(
            dm.filters.Equals(view_id.as_property_ref("standardMarketProductType"), value=standard_market_product_type)
        )
    if standard_market_product_type and isinstance(standard_market_product_type, list):
        filters.append(
            dm.filters.In(view_id.as_property_ref("standardMarketProductType"), values=standard_market_product_type)
        )
    if standard_market_product_type_prefix:
        filters.append(
            dm.filters.Prefix(
                view_id.as_property_ref("standardMarketProductType"), value=standard_market_product_type_prefix
            )
        )
    if original_market_product_type and isinstance(original_market_product_type, str):
        filters.append(
            dm.filters.Equals(view_id.as_property_ref("originalMarketProductType"), value=original_market_product_type)
        )
    if original_market_product_type and isinstance(original_market_product_type, list):
        filters.append(
            dm.filters.In(view_id.as_property_ref("originalMarketProductType"), values=original_market_product_type)
        )
    if original_market_product_type_prefix:
        filters.append(
            dm.filters.Prefix(
                view_id.as_property_ref("originalMarketProductType"), value=original_market_product_type_prefix
            )
        )
    if validity_period and isinstance(validity_period, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("validityPeriod"), value={"space": "power-ops", "externalId": validity_period}
            )
        )
    if validity_period and isinstance(validity_period, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("validityPeriod"),
                value={"space": validity_period[0], "externalId": validity_period[1]},
            )
        )
    if validity_period and isinstance(validity_period, list) and isinstance(validity_period[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("validityPeriod"),
                values=[{"space": "power-ops", "externalId": item} for item in validity_period],
            )
        )
    if validity_period and isinstance(validity_period, list) and isinstance(validity_period[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("validityPeriod"),
                values=[{"space": item[0], "externalId": item[1]} for item in validity_period],
            )
        )
    if reason and isinstance(reason, str):
        filters.append(
            dm.filters.Equals(view_id.as_property_ref("reason"), value={"space": "power-ops", "externalId": reason})
        )
    if reason and isinstance(reason, tuple):
        filters.append(
            dm.filters.Equals(view_id.as_property_ref("reason"), value={"space": reason[0], "externalId": reason[1]})
        )
    if reason and isinstance(reason, list) and isinstance(reason[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("reason"),
                values=[{"space": "power-ops", "externalId": item} for item in reason],
            )
        )
    if reason and isinstance(reason, list) and isinstance(reason[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("reason"), values=[{"space": item[0], "externalId": item[1]} for item in reason]
            )
        )
    if external_id_prefix:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



"""
This module contains the main functions for the resync tool.
"""
from __future__ import annotations

import itertools
import logging
from pathlib import Path
from typing import Literal, Optional, cast

from cognite.client import CogniteClient
from cognite.client.data_classes.data_modeling import DataModelId, MappedProperty, SpaceApply, SpaceList, ViewList
from cognite.client.exceptions import CogniteAPIError
from yaml import safe_dump

from cognite.powerops.cdf_labels import AssetLabel, RelationshipLabel
from cognite.powerops.client.powerops_client import PowerOpsClient
from cognite.powerops.resync import diff, models
from cognite.powerops.resync.config import ReSyncConfig
from cognite.powerops.resync.diff import FieldDifference, ModelDifference, ModelDifferences
from cognite.powerops.resync.models.base import AssetModel, CDFFile, CDFSequence, DataModel, Model, SpaceId
from cognite.powerops.resync.validation import ValidationResults, perform_validation, prepare_validation

from .cdf import get_cognite_api
from .transform import transform
from .validation import _clean_relationships

logger = logging.getLogger(__name__)

MODELS_BY_NAME = {m.__name__: m for m in models.V1_MODELS}

V2_MODELS_BY_NAME = {m.__name__: m for m in models.V2_MODELS}

DATAMODEL_ID_TO_RESYNC_NAME: dict[DataModelId, str] = {
    m.data_model_ids()[0]: m.__name__
    for m in itertools.chain(models.V1_MODELS, models.V2_MODELS)
    if issubclass(m, DataModel)
}


def init(client: PowerOpsClient | None, model_names: str | list[str] | None = None) -> list[dict[str, str]]:
    """
    This function will create the data models in CDF that are required for resync to work. It will not overwrite
    existing models.

    Args:
        client: The PowerOpsClient to use. If not provided, a new client will be created.
        model_names: The models to deploy. If not provided, all models will be deployed.

    """
    client = client or PowerOpsClient.from_settings()
    cdf = client.cdf
    model_classes = _to_models(model_names)

    data_models = [model for model in model_classes if issubclass(model, DataModel)]

    spaces = {space for model in data_models for space in model.spaces()}
    existing_spaces = set((cdf.data_modeling.spaces.retrieve(list(spaces)) or SpaceList([])).as_ids())

    if new_spaces := spaces - existing_spaces:
        logger.info(f"Creating {len(new_spaces)} new spaces: {new_spaces}")
        cdf.data_modeling.spaces.apply(
            [SpaceApply(space, description="PowerOps Configuration Space", name=space.title()) for space in new_spaces]
        )
        logger.info(f"Spaces {new_spaces} created")

    new_data_model_ids = [model_id for m in data_models for model_id in m.data_model_ids()]

    if new_data_model_ids:
        existing = set(cdf.data_modeling.data_models.retrieve(new_data_model_ids).as_ids())
    else:
        existing = set()

    results: list[dict[str, str]] = []
    for model in data_models:
        if model.graph_ql is not None:
            graphql = model.graph_ql
            if graphql.id_ in existing:
                logger.warning(f"Skipping {model.name} data model with {model.graph_ql.id_}, is already exists")
                results.append(
                    {"model": model.name, "action": "skipped"}  # type: ignore[dict-item]
                )  # Ref https://github.com/python/mypy/issues/1465
                continue
            logger.info(f"Deploying {model.name} data model with {model.graph_ql.id_}")
            result = cdf.data_modeling.graphql.apply_dml(
                model.graph_ql.id_, model.graph_ql.graphql, model.graph_ql.name, model.graph_ql.description
            )
            results.append({"model": model.name, "action": "deployed"})  # type: ignore[dict-item]
            logger.info(f"Deployed {model.name} model ({result.space}, {result.external_id}, {result.version})")
        if model.source_model is not None:
            existing_containers = set(
                cdf.data_modeling.containers.retrieve(model.source_model.containers.as_ids()).as_ids()
            )
            new_containers = [
                container for container in model.source_model.containers if container.as_id() not in existing_containers
            ]
            if new_containers:
                logger.info(f"Creating {len(new_containers)} new containers for {model.name}: {new_containers}")
                cdf.data_modeling.containers.apply(new_containers)
                logger.info(f"Containers {new_containers} created")
        if model.dms_model is not None:
            existing_views = set(cdf.data_modeling.views.retrieve(model.dms_model.views.as_ids()).as_ids())
            new_views = [view for view in model.dms_model.views if view.as_id() not in existing_views]
            if new_views:
                logger.info(f"Creating {len(new_views)} new views for {model.name}: {new_views}")
                cdf.data_modeling.views.apply(new_views)
                logger.info(f"Views {new_views} created")
            if model.dms_model.id_ in existing:
                logger.warning(f"Skipping {model.name} data model with {model.dms_model.id_}, is already exists")
                results.append({"model": model.name, "action": "skipped"})  # type: ignore[dict-item]
                continue
            logger.info(f"Deploying {model.name} data model with {model.dms_model.id_}")
            cdf.data_modeling.data_models.apply(model.dms_model.data_model)
            results.append({"model": model.name, "action": "deployed"})  # type: ignore[dict-item]
    return results


def validate(config_dir: str | Path, market: str) -> ValidationResults:
    """
    Validates the local configuration files.

    Args:
        config_dir: Local path to the configuration files. Needs to follow a specific structure. See below.
        market: The market to load the configuration for.

    Configuration file structure:
    ```
    📦config_dir
     ┣ 📂cogshop - The CogSHOP configuration
     ┣ 📂market - The Market configuration for DayAhead, RKOM, and benchmarking.
     ┣ 📂production - The physical assets configuration, Watercourse, PriceArea, Genertor, Plant  (SHOP centered)
     ┗ 📜settings.yaml - Settings for resync.
    ```
    """
    market = market.lower()
    po_client = PowerOpsClient.from_settings()
    logger.info(f"Validating configuration in {config_dir}..")
    loaded_models = _load_transform(market, Path(config_dir), po_client.cdf.config.project, list(MODELS_BY_NAME))

    logger.info("Validating time series...")
    ts_validations, validation_ranges = prepare_validation(loaded_models)
    validation_results = perform_validation(po_client, ts_validations, validation_ranges)
    logger.info("Validations complete")
    return validation_results


def plan(
    config_dir: Path,
    market: str,
    client: PowerOpsClient | None,
    model_names: str | list[str] | None = None,
    dump_folder: Optional[Path] = None,
) -> ModelDifferences:
    """
    Loads the local configuration files, transform them into Resync models, and compares them to the downloaded
    CDF Resync models.

    Args:
        config_dir: Local path to the configuration files. Needs to follow a specific structure. See below.
        market: The market to load the configuration for.
        client: The PowerOpsClient to use. If not provided, a new client will be created.
        model_names: The models to run the plan.
        dump_folder: If present, the local and CDF changes will be dumped to this directory. This is done so that
                    you can use local tools (for example, PyCharm or VS Code) to compare the detailed changes
                    between the local and CDF configuration.


    Returns:
        A ModelDifferences object containing the differences between the local and CDF configuration.

    Configuration file structure:
    ```
    📦config_dir
     ┣ 📂cogshop - The CogSHOP configuration
     ┣ 📂market - The Market configuration for DayAhead, RKOM, and benchmarking.
     ┣ 📂production - The physical assets configuration, Watercourse, PriceArea, Genertor, Plant  (SHOP centered)
     ┗ 📜settings.yaml - Settings for resync.
    ```
    """
    client = client or PowerOpsClient.from_settings()

    loaded_models = _load_transform(market, config_dir, client.cdf.config.project, model_names)

    logger.info(f"Load transform completed, models {', '.join([type(m).__name__ for m in loaded_models])} loaded")
    if client.datasets.read_dataset is None:
        raise ValueError("No read_dataset configured in settings")
    data_set_external_id = client.datasets.read_dataset
    all_differences = []
    for new_model in loaded_models:
        logger.info(f"Retrieving {new_model.model_name} from CDF")
        cdf_model = type(new_model).from_cdf(client, data_set_external_id=data_set_external_id)

        if isinstance(new_model, AssetModel):
            static_resources = new_model.static_resources_from_cdf(client)
        else:
            static_resources = {}

        differences = diff.model_difference(cdf_model, new_model, static_resources)
        _clean_relationships(client.cdf, differences, new_model)

        if dump_folder:
            dump_folder.mkdir(parents=True, exist_ok=True)
            # Standardize models for easy comparison
            new_model.standardize()
            cdf_model.standardize()

            (dump_folder / f"{new_model.model_name}_local.yaml").write_text(safe_dump(new_model.dump_as_cdf_resource()))
            (dump_folder / f"{new_model.model_name}_cdf.yaml").write_text(safe_dump(cdf_model.dump_as_cdf_resource()))

        all_differences.append(differences)
    return ModelDifferences(all_differences)


def apply(
    config_dir: Path,
    market: str,
    client: PowerOpsClient | None = None,
    model_names: str | list[str] | None = None,
    auto_yes: bool = False,
) -> ModelDifferences:
    """
    Loads the local configuration files, transform them into Resync models, and uploads it to CDF. Any deviations
    of the existing CDF configuration will be overwritten.

    Args:
        config_dir: Local path to the configuration files. Needs to follow a specific structure. See below.
        market: The market to load the configuration for.
        client: The PowerOpsClient to use. If not provided, a new client will be created.
        model_names: The models to run the plan.
        auto_yes: If true, all prompts will be auto confirmed.

    Returns:
        A ModelDifferences object containing the differences between the local and CDF configuration which have been
        written to CDF.

    Configuration file structure:
    ```
    📦config_dir
     ┣ 📂cogshop - The CogSHOP configuration
     ┣ 📂market - The Market configuration for DayAhead, RKOM, and benchmarking.
     ┣ 📂production - The physical assets configuration, Watercourse, PriceArea, Genertor, Plant  (SHOP centered)
     ┗ 📜settings.yaml - Settings for resync.
    ```
    """
    client = client or PowerOpsClient.from_settings()
    loaded_models = _load_transform(market, config_dir, client.cdf.config.project, model_names)

    written_changes = ModelDifferences([])
    for new_model in loaded_models:
        cdf_model = type(new_model).from_cdf(client, data_set_external_id=client.datasets.read_dataset)
        if isinstance(new_model, AssetModel):
            static_resources = new_model.static_resources_from_cdf(client)
        else:
            static_resources = {}

        differences = diff.model_difference(cdf_model, new_model, static_resources)

        # Do not create relationships to time series that does not exist.
        _clean_relationships(client.cdf, differences, new_model)
        # Remove the domain model as this is not CDF resources
        # TimeSeries are not updated by resync.
        differences.filter_out(group="Domain", field_names={"timeseries"})

        new_sequences_by_id = {s.external_id: s for s in new_model.sequences()}
        new_files_by_id = {f.external_id: f for f in new_model.files()}

        removed = _remove_resources(differences, client.cdf, auto_yes)
        added_updated = _add_update_resources(differences, client, auto_yes, new_sequences_by_id, new_files_by_id)

        written_changes.append(removed + added_updated)

    return written_changes


def destroy(
    client: PowerOpsClient | None,
    model_names: str | list[str] | None = None,
    auto_yes: bool = False,
    dry_run: bool = False,
) -> ModelDifferences:
    """
    Destroys all resync models in CDF. This will also delete all data in the models.

    Args:
        client: The PowerOpsClient to use. If not provided, a new client will be created.
        model_names: The models to destroy.
        auto_yes: If true, all prompts will be auto confirmed.
        dry_run: If true, the models will not be deleted, but the changes will be printed.

    Returns:
        A ModelDifferences object containing the resources that has been destroyed.
    """
    client = client or PowerOpsClient.from_settings()

    model_types = _to_models(model_names)
    destroyed = ModelDifferences([])
    for model_type in model_types:
        if issubclass(model_type, DataModel) and model_type.graph_ql:
            remove_data_model = _get_data_model_view_containers(
                client.cdf, model_type.graph_ql.id_, model_type.__name__
            )
            if not remove_data_model.changes:
                logger.warning(f"Skipping {model_type.__name__}, no data model found")
                continue
            static_resources = {}
        elif issubclass(model_type, DataModel) and model_type.dms_model:
            raise NotImplementedError()
        elif issubclass(model_type, DataModel) and model_type.source_model:
            raise NotImplementedError()
        elif issubclass(model_type, AssetModel):
            remove_data_model = ModelDifference(model_type.__name__, {})
            if issubclass(model_type, models.MarketModel):
                # We only need the root asset to be set.
                production_external_id = cast(str, models.ProductionModel.root_asset.external_id)
                model_type.set_root_asset("", "", "", production_external_id)
            static_resources = model_type.static_resources_from_cdf(client)
        else:
            raise ValueError(f"Unknown model type {model_type}")

        cdf_model = model_type.from_cdf(client, data_set_external_id=client.datasets.read_dataset)

        remove_data = diff.remove_only(cdf_model, static_resources)
        remove_data.filter_out(group="Domain", field_names={"timeseries"})

        if dry_run:
            destroyed.append(remove_data + remove_data_model)
        else:
            removed = _remove_resources(remove_data + remove_data_model, client.cdf, auto_yes)
            destroyed.append(removed)

    # Spaces are deleted last, as they might contain other resources.
    spaces = set(space for d in model_types if issubclass(d, DataModel) for space in d.spaces())
    if spaces and not dry_run:
        deleted_space: list[SpaceId] = []
        # One at a time, in case there are other resources in the space that will prevent deletion.
        for space in spaces:
            logger.info(f"Deleting space {space}..")
            try:
                client.cdf.data_modeling.spaces.delete(list(spaces))
            except CogniteAPIError as e:
                logger.warning(f"Failed to delete space {space} with error {e}")
            else:
                logger.info(f"... deleted space {space}")
                deleted_space.append(SpaceId(space))
        if deleted_space:
            destroyed.append(
                ModelDifference(
                    model_name="All Models",
                    changes={
                        "spaces": FieldDifference(
                            group="CDF",
                            field_name="spaces",
                            removed=list(deleted_space),
                            added=[],
                            changed=[],
                            unchanged=[],
                        )
                    },
                )
            )
    if not dry_run:
        labels = AssetLabel.as_label_definitions() + RelationshipLabel.as_label_definitions()
        client.cdf.labels.delete([label.external_id for label in labels if label.external_id])

    return destroyed


def migration(client: PowerOpsClient | None, model: Literal["Production"] = "Production") -> ModelDifferences:
    if isinstance(model, list):
        model = model[0]
    if model != "Production":
        raise ValueError(f"Unknown model {model}")
    client = client or PowerOpsClient.from_settings()

    production_dm = models.v2.ProductionModelDM.from_cdf(client, data_set_external_id=client.datasets.read_dataset)
    production_dm_as_asset = models.migration.production_as_asset(production_dm)
    logger.info("Retrieved data model")
    production_asset = models.v1.ProductionModel.from_cdf(client, data_set_external_id=client.datasets.read_dataset)
    logger.info("Retrieved asset model")
    return ModelDifferences([diff.model_difference(current_model=production_asset, new_model=production_dm_as_asset)])


def _remove_resources(differences: ModelDifference, cdf: CogniteClient, auto_yes: bool) -> ModelDifference:
    removed = ModelDifference(model_name=differences.model_name, changes={})
    for difference in sorted(differences, key=_remove_order):
        if not difference.removed:
            continue
        logger.info(f"Removals detected for {difference.field_name} in {differences.model_name}")
        logger.info(f"Remove count: {difference.as_summary().removed}")
        diff_ids = difference.as_ids(5)
        logger.info(f"Sample removed for {diff_ids.field_name}: {diff_ids.removed}")
        ans = "y" if auto_yes else input("Continue? (y/n)")
        if ans.lower() != "y":
            logger.info("Aborting")
            continue

        api = get_cognite_api(cdf, difference.field_name)

        api.delete(external_id=difference.removed_ids)

        logger.info(f"Deleted {len(difference.removed)} of {difference.field_name}")
        removed.changes[difference.field_name] = FieldDifference(
            group=difference.group,
            field_name=difference.field_name,
            added=[],
            removed=difference.removed,
            changed=[],
            unchanged=difference.unchanged,
        )
    return removed


def _add_update_resources(
    differences: ModelDifference,
    client: PowerOpsClient,
    auto_yes: bool,
    sequences_by_id: dict[str, CDFSequence],
    files_by_id: dict[str, CDFFile],
) -> ModelDifference:
    added_updated = ModelDifference(model_name=differences.model_name, changes={})
    for difference in sorted(differences, key=_adding_order):
        if not difference.added and not difference.changed:
            continue
        logger.info(f"Changes/Additions detected for {difference.field_name} in {differences.model_name}")
        summary_count = difference.as_summary()
        logger.info(f"Change count: {summary_count.changed}")
        logger.info(f"Addition count: {summary_count.added}")
        diff_ids = difference.as_ids(5)
        if diff_ids.changed:
            logger.info(f"Sample changed for {diff_ids.field_name}:")
            for change in difference.changed[:3]:
                logger.info(change.changed_fields)
        if diff_ids.added:
            logger.info(f"Sample added for {diff_ids.field_name}: {diff_ids.added}")
        ans = "y" if auto_yes else input("Continue? (y/n)")
        if ans.lower() != "y":
            logger.info("Aborting")
            continue

        difference.set_set_dataset(client.datasets.write_dataset_id)
        api = get_cognite_api(client.cdf, difference.field_name, sequences_by_id, files_by_id)

        if difference.added:
            api.create(difference.added)
            logger.info(f"Created {len(difference.added)} of {difference.field_name}")
            if difference.field_name in added_updated:
                added_updated[difference.field_name].added.extend(difference.added)
            else:
                added_updated[difference.field_name] = FieldDifference(
                    group=difference.group,
                    field_name=difference.field_name,
                    added=difference.added,
                    removed=[],
                    changed=[],
                    unchanged=[],
                )

        if difference.changed:
            updates = [c.new for c in difference.changed if not c.is_changed_content]
            if updates:
                api.upsert(updates, mode="replace")
                logger.info(f"Updated {len(updates)} of {difference.field_name}")
            content_updates = [c.new for c in difference.changed if c.is_changed_content]
            if content_updates:
                api.delete([c.external_id for c in content_updates if c.external_id])
                api.create(content_updates)
                logger.info(f"Updated {len(content_updates)} of {difference.field_name} with content")
            if difference.field_name in added_updated:
                added_updated[difference.field_name].changed.extend(difference.changed)
            else:
                added_updated[difference.field_name] = FieldDifference(
                    group=difference.group,
                    field_name=difference.field_name,
                    added=[],
                    removed=[],
                    changed=difference.changed,
                    unchanged=[],
                )
    return added_updated


def _remove_order(diff: FieldDifference) -> int:
    if diff.group == "Domain":
        return 0
    elif diff.group == "CDF" and diff.field_name not in {
        "edges",
        "nodes",
        "containers",
        "views",
        "data_models",
        "spaces",
        "parent_assets",
        "labels",
    }:
        return 1
    elif diff.group == "CDF" and diff.field_name == "edges":
        return 2
    elif diff.group == "CDF" and diff.field_name == "nodes":
        return 3
    elif diff.group == "CDF" and diff.field_name == "data_models":
        return 4
    elif diff.group == "CDF" and diff.field_name == "views":
        return 5
    elif diff.group == "CDF" and diff.field_name == "containers":
        return 6
    elif diff.group == "CDF" and diff.field_name == "spaces":
        return 7
    elif diff.group == "CDF" and diff.field_name == "parent_assets":
        return 8
    elif diff.group == "CDF" and diff.field_name == "labels":
        return 9
    else:
        return 8


def _adding_order(diff: FieldDifference) -> int:
    if diff.group == "Domain":
        return 0
    elif diff.group == "CDF" and diff.field_name in {"parent_assets", "labels"}:
        return 1
    elif diff.group == "CDF" and diff.field_name not in {"edges", "nodes"}:
        return 2
    elif diff.group == "CDF" and diff.field_name == "nodes":
        return 3
    elif diff.group == "CDF" and diff.field_name == "edges":
        return 4
    else:
        return 5


def _load_transform(market: str, path: Path, cdf_project: str, model_names: str | list[str] | None) -> list[Model]:
    model_types = _to_models(model_names)
    logger.info(f"Loading and transforming {', '.join([m.__name__ for m in model_types])}")

    config = ReSyncConfig.from_yamls(path, cdf_project)

    return transform(config, market, set(model_types))


def _to_models(model_names: str | list[str] | None) -> list[type[Model]]:
    if isinstance(model_names, str):
        model_names = [model_names]
    elif model_names is None:
        model_names = list(MODELS_BY_NAME)
    elif isinstance(model_names, list) and model_names and isinstance(model_names[0], str):
        model_names = model_names
    else:
        raise ValueError(f"Invalid model_names type: {type(model_names)}")

    if invalid := set(model_names) - (set(MODELS_BY_NAME) | set(V2_MODELS_BY_NAME)):
        raise ValueError(f"Invalid model names: {invalid}. Available models: {list(MODELS_BY_NAME)}")

    return [
        MODELS_BY_NAME[model_name] if model_name in MODELS_BY_NAME else V2_MODELS_BY_NAME[model_name]
        for model_name in model_names
    ]


def _get_data_model_view_containers(cdf: CogniteClient, data_model_id: DataModelId, model_name: str) -> ModelDifference:
    data_model = cdf.data_modeling.data_models.retrieve(ids=data_model_id, inline_views=True)
    if not data_model:
        return ModelDifference(model_name=model_name, changes={})
    data_model = data_model.latest_version()
    views = data_model.views

    changes = {
        "views": FieldDifference(group="CDF", field_name="views", removed=list(ViewList(views).as_ids())),
        "containers": FieldDifference(
            group="CDF",
            field_name="containers",
            removed=list(
                {
                    prop.container
                    for view in views
                    for prop in view.properties.values()
                    if isinstance(prop, MappedProperty)
                }
            ),
        ),
        "data_models": FieldDifference(group="CDF", field_name="data_models", removed=[data_model.as_id()]),
    }

    return ModelDifference(model_name=model_name, changes=changes)



from __future__ import annotations

from typing import Literal, Optional

from cognite.client import data_modeling as dm
from pydantic import Field

from ._core import DomainModel, DomainModelApply, TypeApplyList, TypeList

__all__ = [
    "BidCurves",
    "BidCurvesApply",
    "BidCurvesList",
    "BidCurvesApplyList",
    "BidCurvesFields",
    "BidCurvesTextFields",
]


BidCurvesTextFields = Literal["reserve_object"]
BidCurvesFields = Literal["reserve_object", "qty"]

_BIDCURVES_PROPERTIES_BY_FIELD = {
    "reserve_object": "ReserveObject",
    "qty": "Qty",
}


class BidCurves(DomainModel):
    space: str = "power-ops"
    reserve_object: Optional[str] = Field(None, alias="ReserveObject")
    qty: Optional[list[float]] = Field(None, alias="Qty")

    def as_apply(self) -> BidCurvesApply:
        return BidCurvesApply(
            external_id=self.external_id,
            reserve_object=self.reserve_object,
            qty=self.qty,
        )


class BidCurvesApply(DomainModelApply):
    space: str = "power-ops"
    reserve_object: Optional[str] = Field(None, alias="ReserveObject")
    qty: Optional[list[float]] = Field(None, alias="Qty")

    def _to_instances_apply(self, cache: set[str]) -> dm.InstancesApply:
        if self.external_id in cache:
            return dm.InstancesApply(dm.NodeApplyList([]), dm.EdgeApplyList([]))

        sources = []
        properties = {}
        if self.reserve_object is not None:
            properties["ReserveObject"] = self.reserve_object
        if self.qty is not None:
            properties["Qty"] = self.qty
        if properties:
            source = dm.NodeOrEdgeData(
                source=dm.ContainerId("power-ops", "BidCurves"),
                properties=properties,
            )
            sources.append(source)
        if sources:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.existing_version,
                sources=sources,
            )
            nodes = [this_node]
        else:
            nodes = []

        edges = []
        cache.add(self.external_id)

        return dm.InstancesApply(dm.NodeApplyList(nodes), dm.EdgeApplyList(edges))


class BidCurvesList(TypeList[BidCurves]):
    _NODE = BidCurves

    def as_apply(self) -> BidCurvesApplyList:
        return BidCurvesApplyList([node.as_apply() for node in self.data])


class BidCurvesApplyList(TypeApplyList[BidCurvesApply]):
    _NODE = BidCurvesApply
